@@ -27,6 +27,7 @@ Contributors:
 #include "mosquitto_internal.h"
 #include "memory_mosq.h"
 #include "messages_mosq.h"
+#include "mqtt_protocol.h"
 #include "net_mosq.h"
 #include "packet_mosq.h"
 #include "will_mosq.h"
@@ -73,12 +74,12 @@ int mosquitto_lib_cleanup(void)
 	return MOSQ_ERR_SUCCESS;
 }
 
-struct mosquitto *mosquitto_new(const char *id, bool clean_session, void *userdata)
+struct mosquitto *mosquitto_new(const char *id, bool clean_start, void *userdata)
 {
 	struct mosquitto *mosq = NULL;
 	int rc;
 
-	if(clean_session == false && id == NULL){
+	if(clean_start == false && id == NULL){
 		errno = EINVAL;
 		return NULL;
 	}
@@ -95,7 +96,7 @@ struct mosquitto *mosquitto_new(const char *id, bool clean_session, void *userda
 #ifdef WITH_THREADING
 		mosq->thread_id = pthread_self();
 #endif
-		rc = mosquitto_reinitialise(mosq, id, clean_session, userdata);
+		rc = mosquitto_reinitialise(mosq, id, clean_start, userdata);
 		if(rc){
 			mosquitto_destroy(mosq);
 			if(rc == MOSQ_ERR_INVAL){
@@ -111,13 +112,11 @@ struct mosquitto *mosquitto_new(const char *id, bool clean_session, void *userda
 	return mosq;
 }
 
-int mosquitto_reinitialise(struct mosquitto *mosq, const char *id, bool clean_session, void *userdata)
+int mosquitto_reinitialise(struct mosquitto *mosq, const char *id, bool clean_start, void *userdata)
 {
-	int i;
-
 	if(!mosq) return MOSQ_ERR_INVAL;
 
-	if(clean_session == false && id == NULL){
+	if(clean_start == false && id == NULL){
 		return MOSQ_ERR_INVAL;
 	}
 
@@ -134,7 +133,7 @@ int mosquitto_reinitialise(struct mosquitto *mosq, const char *id, bool clean_se
 	mosq->sockpairR = INVALID_SOCKET;
 	mosq->sockpairW = INVALID_SOCKET;
 	mosq->keepalive = 60;
-	mosq->clean_session = clean_session;
+	mosq->clean_start = clean_start;
 	if(id){
 		if(STREMPTY(id)){
 			return MOSQ_ERR_INVAL;
@@ -143,20 +142,6 @@ int mosquitto_reinitialise(struct mosquitto *mosq, const char *id, bool clean_se
 			return MOSQ_ERR_MALFORMED_UTF8;
 		}
 		mosq->id = mosquitto__strdup(id);
-	}else{
-		mosq->id = (char *)mosquitto__calloc(24, sizeof(char));
-		if(!mosq->id){
-			return MOSQ_ERR_NOMEM;
-		}
-		mosq->id[0] = 'm';
-		mosq->id[1] = 'o';
-		mosq->id[2] = 's';
-		mosq->id[3] = 'q';
-		mosq->id[4] = '/';
-
-		for(i=5; i<23; i++){
-			mosq->id[i] = (rand()%73)+48;
-		}
 	}
 	mosq->in_packet.payload = NULL;
 	packet__cleanup(&mosq->in_packet);
@@ -167,11 +152,11 @@ int mosquitto_reinitialise(struct mosquitto *mosq, const char *id, bool clean_se
 	mosq->ping_t = 0;
 	mosq->last_mid = 0;
 	mosq->state = mosq_cs_new;
-	mosq->in_messages = NULL;
-	mosq->in_messages_last = NULL;
-	mosq->out_messages = NULL;
-	mosq->out_messages_last = NULL;
-	mosq->max_inflight_messages = 20;
+	mosq->maximum_qos = 2;
+	mosq->msgs_in.inflight_maximum = 20;
+	mosq->msgs_out.inflight_maximum = 20;
+	mosq->msgs_in.inflight_quota = 20;
+	mosq->msgs_out.inflight_quota = 20;
 	mosq->will = NULL;
 	mosq->on_connect = NULL;
 	mosq->on_publish = NULL;
@@ -181,8 +166,6 @@ int mosquitto_reinitialise(struct mosquitto *mosq, const char *id, bool clean_se
 	mosq->host = NULL;
 	mosq->port = 1883;
 	mosq->in_callback = false;
-	mosq->in_queue_len = 0;
-	mosq->out_queue_len = 0;
 	mosq->reconnect_delay = 1;
 	mosq->reconnect_delay_max = 1;
 	mosq->reconnect_exponential_backoff = false;
@@ -193,6 +176,7 @@ int mosquitto_reinitialise(struct mosquitto *mosq, const char *id, bool clean_se
 	mosq->tls_cert_reqs = SSL_VERIFY_PEER;
 	mosq->tls_insecure = false;
 	mosq->want_write = false;
+	mosq->tls_ocsp_required = false;
 #endif
 #ifdef WITH_THREADING
 	pthread_mutex_init(&mosq->callback_mutex, NULL);
@@ -201,8 +185,8 @@ int mosquitto_reinitialise(struct mosquitto *mosq, const char *id, bool clean_se
 	pthread_mutex_init(&mosq->out_packet_mutex, NULL);
 	pthread_mutex_init(&mosq->current_out_packet_mutex, NULL);
 	pthread_mutex_init(&mosq->msgtime_mutex, NULL);
-	pthread_mutex_init(&mosq->in_message_mutex, NULL);
-	pthread_mutex_init(&mosq->out_message_mutex, NULL);
+	pthread_mutex_init(&mosq->msgs_in.mutex, NULL);
+	pthread_mutex_init(&mosq->msgs_out.mutex, NULL);
 	pthread_mutex_init(&mosq->mid_mutex, NULL);
 	mosq->thread_id = pthread_self();
 #endif
@@ -233,8 +217,8 @@ void mosquitto__destroy(struct mosquitto *mosq)
 		pthread_mutex_destroy(&mosq->out_packet_mutex);
 		pthread_mutex_destroy(&mosq->current_out_packet_mutex);
 		pthread_mutex_destroy(&mosq->msgtime_mutex);
-		pthread_mutex_destroy(&mosq->in_message_mutex);
-		pthread_mutex_destroy(&mosq->out_message_mutex);
+		pthread_mutex_destroy(&mosq->msgs_in.mutex);
+		pthread_mutex_destroy(&mosq->msgs_out.mutex);
 		pthread_mutex_destroy(&mosq->mid_mutex);
 	}
 #endif
@@ -259,6 +243,7 @@ void mosquitto__destroy(struct mosquitto *mosq)
 	mosquitto__free(mosq->tls_ciphers);
 	mosquitto__free(mosq->tls_psk);
 	mosquitto__free(mosq->tls_psk_identity);
+	mosquitto__free(mosq->tls_alpn);
 #endif
 
 	mosquitto__free(mosq->address);
@@ -344,6 +329,12 @@ bool mosquitto_want_write(struct mosquitto *mosq)
 const char *mosquitto_strerror(int mosq_errno)
 {
 	switch(mosq_errno){
+		case MOSQ_ERR_AUTH_CONTINUE:
+			return "Continue with authentication.";
+		case MOSQ_ERR_NO_SUBSCRIBERS:
+			return "No subscribers.";
+		case MOSQ_ERR_SUB_EXISTS:
+			return "Subscription already exists.";
 		case MOSQ_ERR_CONN_PENDING:
 			return "Connection pending.";
 		case MOSQ_ERR_SUCCESS:
@@ -382,6 +373,16 @@ const char *mosquitto_strerror(int mosq_errno)
 			return "Proxy error.";
 		case MOSQ_ERR_MALFORMED_UTF8:
 			return "Malformed UTF-8";
+		case MOSQ_ERR_DUPLICATE_PROPERTY:
+			return "Duplicate property in property list";
+		case MOSQ_ERR_TLS_HANDSHAKE:
+			return "TLS handshake failed.";
+		case MOSQ_ERR_QOS_NOT_SUPPORTED:
+			return "Requested QoS not supported on server.";
+		case MOSQ_ERR_OVERSIZE_PACKET:
+			return "Packet larger than supported by the server.";
+		case MOSQ_ERR_OCSP:
+			return "OCSP error.";
 		default:
 			return "Unknown error.";
 	}
@@ -407,6 +408,135 @@ const char *mosquitto_connack_string(int connack_code)
 	}
 }
 
+const char *mosquitto_reason_string(int reason_code)
+{
+	switch(reason_code){
+		case MQTT_RC_SUCCESS:
+			return "Success";
+		case MQTT_RC_GRANTED_QOS1:
+			return "Granted QoS 1";
+		case MQTT_RC_GRANTED_QOS2:
+			return "Granted QoS 2";
+		case MQTT_RC_DISCONNECT_WITH_WILL_MSG:
+			return "Disconnect with Will Message";
+		case MQTT_RC_NO_MATCHING_SUBSCRIBERS:
+			return "No matching subscribers";
+		case MQTT_RC_NO_SUBSCRIPTION_EXISTED:
+			return "No subscription existed";
+		case MQTT_RC_CONTINUE_AUTHENTICATION:
+			return "Continue authentication";
+		case MQTT_RC_REAUTHENTICATE:
+			return "Re-authenticate";
+
+		case MQTT_RC_UNSPECIFIED:
+			return "Unspecified error";
+		case MQTT_RC_MALFORMED_PACKET:
+			return "Malformed Packet";
+		case MQTT_RC_PROTOCOL_ERROR:
+			return "Protocol Error";
+		case MQTT_RC_IMPLEMENTATION_SPECIFIC:
+			return "Implementation specific error";
+		case MQTT_RC_UNSUPPORTED_PROTOCOL_VERSION:
+			return "Unsupported Protocol Version";
+		case MQTT_RC_CLIENTID_NOT_VALID:
+			return "Client Identifier not valid";
+		case MQTT_RC_BAD_USERNAME_OR_PASSWORD:
+			return "Bad User Name or Password";
+		case MQTT_RC_NOT_AUTHORIZED:
+			return "Not authorized";
+		case MQTT_RC_SERVER_UNAVAILABLE:
+			return "Server unavailable";
+		case MQTT_RC_SERVER_BUSY:
+			return "Server busy";
+		case MQTT_RC_BANNED:
+			return "Banned";
+		case MQTT_RC_SERVER_SHUTTING_DOWN:
+			return "Server shutting down";
+		case MQTT_RC_BAD_AUTHENTICATION_METHOD:
+			return "Bad authentication method";
+		case MQTT_RC_KEEP_ALIVE_TIMEOUT:
+			return "Keep Alive timeout";
+		case MQTT_RC_SESSION_TAKEN_OVER:
+			return "Session taken over";
+		case MQTT_RC_TOPIC_FILTER_INVALID:
+			return "Topic Filter invalid";
+		case MQTT_RC_TOPIC_NAME_INVALID:
+			return "Topic Name invalid";
+		case MQTT_RC_PACKET_ID_IN_USE:
+			return "Packet Identifier in use";
+		case MQTT_RC_PACKET_ID_NOT_FOUND:
+			return "Packet Identifier not found";
+		case MQTT_RC_RECEIVE_MAXIMUM_EXCEEDED:
+			return "Receive Maximum exceeded";
+		case MQTT_RC_TOPIC_ALIAS_INVALID:
+			return "Topic Alias invalid";
+		case MQTT_RC_PACKET_TOO_LARGE:
+			return "Packet too large";
+		case MQTT_RC_MESSAGE_RATE_TOO_HIGH:
+			return "Message rate too high";
+		case MQTT_RC_QUOTA_EXCEEDED:
+			return "Quota exceeded";
+		case MQTT_RC_ADMINISTRATIVE_ACTION:
+			return "Administrative action";
+		case MQTT_RC_PAYLOAD_FORMAT_INVALID:
+			return "Payload format invalid";
+		case MQTT_RC_RETAIN_NOT_SUPPORTED:
+			return "Retain not supported";
+		case MQTT_RC_QOS_NOT_SUPPORTED:
+			return "QoS not supported";
+		case MQTT_RC_USE_ANOTHER_SERVER:
+			return "Use another server";
+		case MQTT_RC_SERVER_MOVED:
+			return "Server moved";
+		case MQTT_RC_SHARED_SUBS_NOT_SUPPORTED:
+			return "Shared Subscriptions not supported";
+		case MQTT_RC_CONNECTION_RATE_EXCEEDED:
+			return "Connection rate exceeded";
+		case MQTT_RC_MAXIMUM_CONNECT_TIME:
+			return "Maximum connect time";
+		case MQTT_RC_SUBSCRIPTION_IDS_NOT_SUPPORTED:
+			return "Subscription identifiers not supported";
+		case MQTT_RC_WILDCARD_SUBS_NOT_SUPPORTED:
+			return "Wildcard Subscriptions not supported";
+		default:
+			return "Unknown reason";
+	}
+}
+
+
+int mosquitto_string_to_command(const char *str, int *cmd)
+{
+	if(!strcasecmp(str, "connect")){
+		*cmd = CMD_CONNECT;
+	}else if(!strcasecmp(str, "connack")){
+		*cmd = CMD_CONNACK;
+	}else if(!strcasecmp(str, "publish")){
+		*cmd = CMD_PUBLISH;
+	}else if(!strcasecmp(str, "puback")){
+		*cmd = CMD_PUBACK;
+	}else if(!strcasecmp(str, "pubrec")){
+		*cmd = CMD_PUBREC;
+	}else if(!strcasecmp(str, "pubrel")){
+		*cmd = CMD_PUBREL;
+	}else if(!strcasecmp(str, "pubcomp")){
+		*cmd = CMD_PUBCOMP;
+	}else if(!strcasecmp(str, "subscribe")){
+		*cmd = CMD_SUBSCRIBE;
+	}else if(!strcasecmp(str, "unsubscribe")){
+		*cmd = CMD_UNSUBSCRIBE;
+	}else if(!strcasecmp(str, "disconnect")){
+		*cmd = CMD_DISCONNECT;
+	}else if(!strcasecmp(str, "auth")){
+		*cmd = CMD_AUTH;
+	}else if(!strcasecmp(str, "will")){
+		*cmd = CMD_WILL;
+	}else{
+		return MOSQ_ERR_INVAL;
+	}
+	return MOSQ_ERR_SUCCESS;
+}
+
+
 int mosquitto_sub_topic_tokenise(const char *subtopic, char ***topics, int *count)
 {
 	int len;
