@@ -109,13 +109,18 @@ enum mosquitto_client_state {
 	mosq_cs_expiring = 15,
 	mosq_cs_connecting = 16,
 	mosq_cs_duplicate = 17, /* client that has been taken over by another with the same id */
+	mosq_cs_disconnect_with_will = 18,
+	mosq_cs_disused = 19, /* client that has been added to the disused list to be freed */
+	mosq_cs_authenticating = 20, /* Client has sent CONNECT but is still undergoing extended authentication */
+	mosq_cs_reauthenticating = 21, /* Client is undergoing reauthentication and shouldn't do anything else until complete */
 };
 
 enum mosquitto__protocol {
 	mosq_p_invalid = 0,
 	mosq_p_mqtt31 = 1,
 	mosq_p_mqtt311 = 2,
-	mosq_p_mqtts = 3
+	mosq_p_mqtts = 3,
+	mosq_p_mqtt5 = 5,
 };
 
 enum mosquitto__threaded_state {
@@ -131,6 +136,18 @@ enum mosquitto__transport {
 	mosq_t_sctp = 3
 };
 
+
+struct mosquitto__alias{
+	char *topic;
+	uint16_t alias;
+};
+
+struct session_expiry_list {
+	struct mosquitto *context;
+	struct session_expiry_list *prev;
+	struct session_expiry_list *next;
+};
+
 struct mosquitto__packet{
 	uint8_t *payload;
 	struct mosquitto__packet *next;
@@ -146,13 +163,43 @@ struct mosquitto__packet{
 
 struct mosquitto_message_all{
 	struct mosquitto_message_all *next;
+	struct mosquitto_message_all *prev;
+	mosquitto_property *properties;
 	time_t timestamp;
 	//enum mosquitto_msg_direction direction;
 	enum mosquitto_msg_state state;
 	bool dup;
 	struct mosquitto_message msg;
+	uint32_t expiry_interval;
 };
 
+#ifdef WITH_TLS
+enum mosquitto__keyform {
+	mosq_k_pem = 0,
+	mosq_k_engine = 1,
+};
+#endif
+
+struct mosquitto_msg_data{
+#ifdef WITH_BROKER
+	struct mosquitto_client_msg *inflight;
+	struct mosquitto_client_msg *queued;
+	unsigned long msg_bytes;
+	unsigned long msg_bytes12;
+	int msg_count;
+	int msg_count12;
+#else
+	struct mosquitto_message_all *inflight;
+	int queue_len;
+#  ifdef WITH_THREADING
+	pthread_mutex_t mutex;
+#  endif
+#endif
+	int inflight_quota;
+	uint16_t inflight_maximum;
+};
+
+
 struct mosquitto {
 	mosq_sock_t sock;
 #ifndef WITH_BROKER
@@ -175,7 +222,12 @@ struct mosquitto {
 	struct mosquitto__packet in_packet;
 	struct mosquitto__packet *current_out_packet;
 	struct mosquitto__packet *out_packet;
-	struct mosquitto_message *will;
+	struct mosquitto_message_all *will;
+	struct mosquitto__alias *aliases;
+	uint32_t maximum_packet_size;
+	int alias_count;
+	uint32_t will_delay_interval;
+	time_t will_delay_time;
 #ifdef WITH_TLS
 	SSL *ssl;
 	SSL_CTX *ssl_ctx;
@@ -191,6 +243,11 @@ struct mosquitto {
 	int tls_cert_reqs;
 	bool tls_insecure;
 	bool ssl_ctx_defaults;
+	bool tls_ocsp_required;
+	char *tls_engine;
+	char *tls_engine_kpass_sha1;
+	enum mosquitto__keyform tls_keyform;
+	char *tls_alpn;
 #endif
 	bool want_write;
 	bool want_connect;
@@ -201,31 +258,27 @@ struct mosquitto {
 	pthread_mutex_t out_packet_mutex;
 	pthread_mutex_t current_out_packet_mutex;
 	pthread_mutex_t state_mutex;
-	pthread_mutex_t in_message_mutex;
-	pthread_mutex_t out_message_mutex;
 	pthread_mutex_t mid_mutex;
 	pthread_t thread_id;
 #endif
-	bool clean_session;
+	bool clean_start;
+	uint32_t session_expiry_interval;
+	time_t session_expiry_time;
 #ifdef WITH_BROKER
 	bool removed_from_by_id; /* True if removed from by_id hash */
 	bool is_dropping;
 	bool is_bridge;
 	struct mosquitto__bridge *bridge;
-	struct mosquitto_client_msg *inflight_msgs;
-	struct mosquitto_client_msg *last_inflight_msg;
-	struct mosquitto_client_msg *queued_msgs;
-	struct mosquitto_client_msg *last_queued_msg;
-	unsigned long msg_bytes;
-	unsigned long msg_bytes12;
-	int msg_count;
-	int msg_count12;
+	struct mosquitto_msg_data msgs_in;
+	struct mosquitto_msg_data msgs_out;
 	struct mosquitto__acl_user *acl_list;
 	struct mosquitto__listener *listener;
-	time_t disconnect_t;
 	struct mosquitto__packet *out_packet_last;
 	struct mosquitto__subhier **subs;
+	struct mosquitto__subshared_ref **shared_subs;
+	char *auth_method;
 	int sub_count;
+	int shared_sub_count;
 	int pollfd_index;
 #  ifdef WITH_WEBSOCKETS
 #    if defined(LWS_LIBRARY_VERSION_NUMBER)
@@ -236,6 +289,7 @@ struct mosquitto {
 #    endif
 #  endif
 	bool ws_want_write;
+	bool assigned_id;
 #else
 #  ifdef WITH_SOCKS
 	char *socks5_host;
@@ -245,40 +299,44 @@ struct mosquitto {
 #  endif
 	void *userdata;
 	bool in_callback;
-	struct mosquitto_message_all *in_messages;
-	struct mosquitto_message_all *in_messages_last;
-	struct mosquitto_message_all *out_messages;
-	struct mosquitto_message_all *out_messages_last;
+	struct mosquitto_msg_data msgs_in;
+	struct mosquitto_msg_data msgs_out;
 	void (*on_connect)(struct mosquitto *, void *userdata, int rc);
 	void (*on_connect_with_flags)(struct mosquitto *, void *userdata, int rc, int flags);
+	void (*on_connect_v5)(struct mosquitto *, void *userdata, int rc, int flags, const mosquitto_property *props);
 	void (*on_disconnect)(struct mosquitto *, void *userdata, int rc);
+	void (*on_disconnect_v5)(struct mosquitto *, void *userdata, int rc, const mosquitto_property *props);
 	void (*on_publish)(struct mosquitto *, void *userdata, int mid);
+	void (*on_publish_v5)(struct mosquitto *, void *userdata, int mid, int reason_code, const mosquitto_property *props);
 	void (*on_message)(struct mosquitto *, void *userdata, const struct mosquitto_message *message);
+	void (*on_message_v5)(struct mosquitto *, void *userdata, const struct mosquitto_message *message, const mosquitto_property *props);
 	void (*on_subscribe)(struct mosquitto *, void *userdata, int mid, int qos_count, const int *granted_qos);
+	void (*on_subscribe_v5)(struct mosquitto *, void *userdata, int mid, int qos_count, const int *granted_qos, const mosquitto_property *props);
 	void (*on_unsubscribe)(struct mosquitto *, void *userdata, int mid);
+	void (*on_unsubscribe_v5)(struct mosquitto *, void *userdata, int mid, const mosquitto_property *props);
 	void (*on_log)(struct mosquitto *, void *userdata, int level, const char *str);
 	//void (*on_error)();
 	char *host;
 	int port;
-	int in_queue_len;
-	int out_queue_len;
 	char *bind_address;
 	unsigned int reconnect_delay;
 	unsigned int reconnect_delay_max;
 	bool reconnect_exponential_backoff;
 	char threaded;
 	struct mosquitto__packet *out_packet_last;
-	int inflight_messages;
-	int max_inflight_messages;
 #  ifdef WITH_SRV
 	ares_channel achan;
 #  endif
+	uint16_t send_maximum;
+	uint16_t receive_maximum;
 #endif
+	uint8_t maximum_qos;
 
 #ifdef WITH_BROKER
 	UT_hash_handle hh_id;
 	UT_hash_handle hh_sock;
 	struct mosquitto *for_free_next;
+	struct session_expiry_list *expiry_list_item;
 #endif
 #ifdef WITH_EPOLL
 	uint32_t events;
@@ -287,5 +345,7 @@ struct mosquitto {
 
 #define STREMPTY(str) (str[0] == '\0')
 
+void do_client_disconnect(struct mosquitto *mosq, int reason_code, const mosquitto_property *properties);
+
 #endif
 
