@@ -1,5 +1,5 @@
 /*
-Copyright (c) 2010-2018 Roger Light <roger@atchoo.org>
+Copyright (c) 2010-2019 Roger Light <roger@atchoo.org>
 
 All rights reserved. This program and the accompanying materials
 are made available under the terms of the Eclipse Public License v1.0
@@ -27,15 +27,24 @@ int packet__alloc(struct mosquitto__packet *packet);
 void packet__cleanup(struct mosquitto__packet *packet);
 int packet__queue(struct mosquitto *mosq, struct mosquitto__packet *packet);
 
+int packet__check_oversize(struct mosquitto *mosq, uint32_t remaining_length);
+
 int packet__read_byte(struct mosquitto__packet *packet, uint8_t *byte);
 int packet__read_bytes(struct mosquitto__packet *packet, void *bytes, uint32_t count);
+int packet__read_binary(struct mosquitto__packet *packet, uint8_t **data, int *length);
 int packet__read_string(struct mosquitto__packet *packet, char **str, int *length);
 int packet__read_uint16(struct mosquitto__packet *packet, uint16_t *word);
+int packet__read_uint32(struct mosquitto__packet *packet, uint32_t *word);
+int packet__read_varint(struct mosquitto__packet *packet, int32_t *word, int8_t *bytes);
 
 void packet__write_byte(struct mosquitto__packet *packet, uint8_t byte);
 void packet__write_bytes(struct mosquitto__packet *packet, const void *bytes, uint32_t count);
 void packet__write_string(struct mosquitto__packet *packet, const char *str, uint16_t length);
 void packet__write_uint16(struct mosquitto__packet *packet, uint16_t word);
+void packet__write_uint32(struct mosquitto__packet *packet, uint32_t word);
+int packet__write_varint(struct mosquitto__packet *packet, int32_t word);
+
+int packet__varint_bytes(int32_t word);
 
 int packet__write(struct mosquitto *mosq);
 #ifdef WITH_BROKER
