@@ -1,5 +1,5 @@
 /*
-Copyright (c) 2012-2018 Roger Light <roger@atchoo.org>
+Copyright (c) 2012-2019 Roger Light <roger@atchoo.org>
 
 All rights reserved. This program and the accompanying materials
 are made available under the terms of the Eclipse Public License v1.0
@@ -17,6 +17,7 @@ Contributors:
 #include "config.h"
 
 #include <errno.h>
+#include <openssl/opensslv.h>
 #include <openssl/evp.h>
 #include <openssl/rand.h>
 #include <openssl/buffer.h>
@@ -45,11 +46,52 @@ Contributors:
 #  include <stdbool.h>
 #  include <unistd.h>
 #  include <termios.h>
+#  include <sys/stat.h>
 #endif
 
 #define MAX_BUFFER_LEN 1024
 #define SALT_LEN 12
 
+#ifdef WIN32
+static FILE *mpw_tmpfile(void)
+{
+	return tmpfile();
+}
+#else
+
+static char alphanum[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
+
+static unsigned char tmpfile_path[36];
+static FILE *mpw_tmpfile(void)
+{
+	int fd;
+	size_t i;
+
+	if(RAND_bytes(tmpfile_path, sizeof(tmpfile_path)) != 1){
+		return NULL;
+	}
+
+	strcpy((char *)tmpfile_path, "/tmp/");
+
+	for(i=strlen((char *)tmpfile_path); i<sizeof(tmpfile_path)-8; i++){
+		tmpfile_path[i] = alphanum[tmpfile_path[i]%(sizeof(alphanum)-1)];
+	}
+	tmpfile_path[sizeof(tmpfile_path)-8] = '-';
+	for(i=sizeof(tmpfile_path)-7; i<sizeof(tmpfile_path)-1; i++){
+		tmpfile_path[i] = 'X';
+	}
+	tmpfile_path[sizeof(tmpfile_path)-1] = '\0';
+
+	umask(077);
+	fd = mkstemp((char *)tmpfile_path);
+	if(fd < 0) return NULL;
+	unlink((char *)tmpfile_path);
+
+	return fdopen(fd, "w+");
+}
+#endif
+
+
 int base64_encode(unsigned char *in, unsigned int in_len, char **encoded)
 {
 	BIO *bmem, *b64;
@@ -88,7 +130,7 @@ void print_usage(void)
 	printf(" -c : create a new password file. This will overwrite existing files.\n");
 	printf(" -D : delete the username rather than adding/updating its password.\n");
 	printf(" -U : update a plain text password file to use hashed passwords.\n");
-	printf("\nSee http://mosquitto.org/ for more information.\n\n");
+	printf("\nSee https://mosquitto.org/ for more information.\n\n");
 }
 
 int output_new_password(FILE *fptr, const char *username, const char *password)
@@ -162,10 +204,17 @@ int delete_pwuser(FILE *fptr, FILE *ftmp, const char *username)
 	char buf[MAX_BUFFER_LEN];
 	char lbuf[MAX_BUFFER_LEN], *token;
 	bool found = false;
+	int line = 0;
 
 	while(!feof(fptr) && fgets(buf, MAX_BUFFER_LEN, fptr)){
+		line++;
 		memcpy(lbuf, buf, MAX_BUFFER_LEN);
 		token = strtok(lbuf, ":");
+		if(!token){
+			fprintf(stderr, "Error: Corrupt password file at line %d.\n", line);
+			return 1;
+		}
+
 		if(strcmp(username, token)){
 			fprintf(ftmp, "%s", buf);
 		}else{
@@ -211,10 +260,17 @@ int update_pwuser(FILE *fptr, FILE *ftmp, const char *username, const char *pass
 	char lbuf[MAX_BUFFER_LEN], *token;
 	bool found = false;
 	int rc = 1;
+	int line = 0;
 
 	while(!feof(fptr) && fgets(buf, MAX_BUFFER_LEN, fptr)){
+		line++;
 		memcpy(lbuf, buf, MAX_BUFFER_LEN);
 		token = strtok(lbuf, ":");
+		if(!token){
+			fprintf(stderr, "Error: Corrupt password file at line %d.\n", line);
+			return 1;
+		}
+
 		if(strcmp(username, token)){
 			fprintf(ftmp, "%s", buf);
 		}else{
@@ -315,7 +371,7 @@ int get_password(char *password, int len)
 int copy_contents(FILE *src, FILE *dest)
 {
 	char buf[MAX_BUFFER_LEN];
-	int len;
+	size_t len;
 
 	rewind(src);
 	rewind(dest);
@@ -366,6 +422,9 @@ void handle_sigint(int signal)
 	ts.c_lflag |= ECHO | ICANON;
 	tcsetattr(0, TCSANOW, &ts);
 #endif
+
+	UNUSED(signal);
+
 	exit(0);
 }
 
@@ -389,6 +448,10 @@ int main(int argc, char *argv[])
 
 #if OPENSSL_VERSION_NUMBER < 0x10100000L || OPENSSL_API_COMPAT < 0x10100000L
 	OpenSSL_add_all_digests();
+#else
+	OPENSSL_init_crypto(OPENSSL_INIT_ADD_ALL_CIPHERS \
+			| OPENSSL_INIT_ADD_ALL_DIGESTS \
+			| OPENSSL_INIT_LOAD_CONFIG, NULL);
 #endif
 
 	if(argc == 1){
@@ -464,7 +527,10 @@ int main(int argc, char *argv[])
 
 	if(create_new){
 		rc = get_password(password, 1024);
-		if(rc) return rc;
+		if(rc){
+			free(password_file);
+			return rc;
+		}
 		fptr = fopen(password_file, "wt");
 		if(!fptr){
 			fprintf(stderr, "Error: Unable to open file %s for writing. %s.\n", password_file, strerror(errno));
@@ -499,7 +565,7 @@ int main(int argc, char *argv[])
 			return 1;
 		}
 
-		ftmp = tmpfile();
+		ftmp = mpw_tmpfile();
 		if(!ftmp){
 			fprintf(stderr, "Error: Unable to open temporary file. %s.\n", strerror(errno));
 			fclose(fptr);
