@@ -1,5 +1,5 @@
 /*
-Copyright (c) 2011-2018 Roger Light <roger@atchoo.org>
+Copyright (c) 2011-2019 Roger Light <roger@atchoo.org>
 
 All rights reserved. This program and the accompanying materials
 are made available under the terms of the Eclipse Public License v1.0
@@ -42,7 +42,7 @@ void LIB_ERROR(void)
 }
 
 
-int security__load_v2(struct mosquitto_db *db, struct mosquitto__auth_plugin *plugin, struct mosquitto_auth_opt *auth_options, int auth_option_count, void *lib)
+int security__load_v2(struct mosquitto__auth_plugin *plugin, struct mosquitto_auth_opt *auth_options, int auth_option_count, void *lib)
 {
 	int rc;
 
@@ -116,7 +116,7 @@ int security__load_v2(struct mosquitto_db *db, struct mosquitto__auth_plugin *pl
 }
 
 
-int security__load_v3(struct mosquitto_db *db, struct mosquitto__auth_plugin *plugin, struct mosquitto_opt *auth_options, int auth_option_count, void *lib)
+int security__load_v3(struct mosquitto__auth_plugin *plugin, struct mosquitto_opt *auth_options, int auth_option_count, void *lib)
 {
 	int rc;
 
@@ -189,7 +189,100 @@ int security__load_v3(struct mosquitto_db *db, struct mosquitto__auth_plugin *pl
 }
 
 
-static int security__module_init_single(struct mosquitto_db *db, struct mosquitto__security_options *opts)
+int security__load_v4(struct mosquitto__auth_plugin *plugin, struct mosquitto_opt *auth_options, int auth_option_count, void *lib)
+{
+	int rc;
+
+	if(!(plugin->plugin_init_v4 = (FUNC_auth_plugin_init_v4)LIB_SYM(lib, "mosquitto_auth_plugin_init"))){
+		log__printf(NULL, MOSQ_LOG_ERR,
+				"Error: Unable to load auth plugin function mosquitto_auth_plugin_init().");
+		LIB_ERROR();
+		LIB_CLOSE(lib);
+		return 1;
+	}
+	if(!(plugin->plugin_cleanup_v4 = (FUNC_auth_plugin_cleanup_v4)LIB_SYM(lib, "mosquitto_auth_plugin_cleanup"))){
+		log__printf(NULL, MOSQ_LOG_ERR,
+				"Error: Unable to load auth plugin function mosquitto_auth_plugin_cleanup().");
+		LIB_ERROR();
+		LIB_CLOSE(lib);
+		return 1;
+	}
+
+	if(!(plugin->security_init_v4 = (FUNC_auth_plugin_security_init_v4)LIB_SYM(lib, "mosquitto_auth_security_init"))){
+		log__printf(NULL, MOSQ_LOG_ERR,
+				"Error: Unable to load auth plugin function mosquitto_auth_security_init().");
+		LIB_ERROR();
+		LIB_CLOSE(lib);
+		return 1;
+	}
+
+	if(!(plugin->security_cleanup_v4 = (FUNC_auth_plugin_security_cleanup_v4)LIB_SYM(lib, "mosquitto_auth_security_cleanup"))){
+		log__printf(NULL, MOSQ_LOG_ERR,
+				"Error: Unable to load auth plugin function mosquitto_auth_security_cleanup().");
+		LIB_ERROR();
+		LIB_CLOSE(lib);
+		return 1;
+	}
+
+	if(!(plugin->acl_check_v4 = (FUNC_auth_plugin_acl_check_v4)LIB_SYM(lib, "mosquitto_auth_acl_check"))){
+		log__printf(NULL, MOSQ_LOG_ERR,
+				"Error: Unable to load auth plugin function mosquitto_auth_acl_check().");
+		LIB_ERROR();
+		LIB_CLOSE(lib);
+		return 1;
+	}
+
+	plugin->unpwd_check_v4 = (FUNC_auth_plugin_unpwd_check_v4)LIB_SYM(lib, "mosquitto_auth_unpwd_check");
+	if(plugin->unpwd_check_v4){
+		log__printf(NULL, MOSQ_LOG_INFO,
+				" ├── Username/password checking enabled.");
+	}else{
+		log__printf(NULL, MOSQ_LOG_INFO,
+				" ├── Username/password checking not enabled.");
+	}
+
+	plugin->psk_key_get_v4 = (FUNC_auth_plugin_psk_key_get_v4)LIB_SYM(lib, "mosquitto_auth_psk_key_get");
+	if(plugin->psk_key_get_v4){
+		log__printf(NULL, MOSQ_LOG_INFO,
+				" ├── TLS-PSK checking enabled.");
+	}else{
+		log__printf(NULL, MOSQ_LOG_INFO,
+				" ├── TLS-PSK checking not enabled.");
+	}
+
+	plugin->auth_start_v4 = (FUNC_auth_plugin_auth_start_v4)LIB_SYM(lib, "mosquitto_auth_start");
+	plugin->auth_continue_v4 = (FUNC_auth_plugin_auth_continue_v4)LIB_SYM(lib, "mosquitto_auth_continue");
+	
+	if(plugin->auth_start_v4){
+		if(plugin->auth_continue_v4){
+			log__printf(NULL, MOSQ_LOG_INFO,
+					" └── Extended authentication enabled.");
+		}else{
+			log__printf(NULL, MOSQ_LOG_ERR,
+					"Error: Plugin has missing mosquitto_auth_continue() function.");
+			LIB_CLOSE(lib);
+			return 1;
+		}
+	}else{
+		log__printf(NULL, MOSQ_LOG_INFO,
+				" └── Extended authentication not enabled.");
+	}
+
+	plugin->lib = lib;
+	plugin->user_data = NULL;
+	if(plugin->plugin_init_v4){
+		rc = plugin->plugin_init_v4(&plugin->user_data, auth_options, auth_option_count);
+		if(rc){
+			log__printf(NULL, MOSQ_LOG_ERR,
+					"Error: Authentication plugin returned %d when initialising.", rc);
+			return rc;
+		}
+	}
+	return 0;
+}
+
+
+static int security__module_init_single(struct mosquitto__security_options *opts)
 {
 	void *lib;
 	int (*plugin_version)(void) = NULL;
@@ -205,6 +298,8 @@ static int security__module_init_single(struct mosquitto_db *db, struct mosquitt
 		if(opts->auth_plugin_configs[i].path){
 			memset(&opts->auth_plugin_configs[i].plugin, 0, sizeof(struct mosquitto__auth_plugin));
 
+			log__printf(NULL, MOSQ_LOG_INFO, "Loading plugin: %s", opts->auth_plugin_configs[i].path);
+
 			lib = LIB_LOAD(opts->auth_plugin_configs[i].path);
 			if(!lib){
 				log__printf(NULL, MOSQ_LOG_ERR,
@@ -223,9 +318,18 @@ static int security__module_init_single(struct mosquitto_db *db, struct mosquitt
 			}
 			version = plugin_version();
 			opts->auth_plugin_configs[i].plugin.version = version;
-			if(version == 3){
+			if(version == 4){
+				rc = security__load_v4(
+						&opts->auth_plugin_configs[i].plugin,
+						opts->auth_plugin_configs[i].options,
+						opts->auth_plugin_configs[i].option_count,
+						lib);
+
+				if(rc){
+					return rc;
+				}
+			}else if(version == 3){
 				rc = security__load_v3(
-						db,
 						&opts->auth_plugin_configs[i].plugin,
 						opts->auth_plugin_configs[i].options,
 						opts->auth_plugin_configs[i].option_count,
@@ -236,7 +340,6 @@ static int security__module_init_single(struct mosquitto_db *db, struct mosquitt
 				}
 			}else if(version == 2){
 				rc = security__load_v2(
-						db,
 						&opts->auth_plugin_configs[i].plugin,
 						(struct mosquitto_auth_opt *)opts->auth_plugin_configs[i].options,
 						opts->auth_plugin_configs[i].option_count,
@@ -267,11 +370,11 @@ int mosquitto_security_module_init(struct mosquitto_db *db)
 
 	if(db->config->per_listener_settings){
 		for(i=0; i<db->config->listener_count; i++){
-			rc = security__module_init_single(db, &db->config->listeners[i].security_options);
+			rc = security__module_init_single(&db->config->listeners[i].security_options);
 			if(rc) return rc;
 		}
 	}else{
-		rc = security__module_init_single(db, &db->config->security_options);
+		rc = security__module_init_single(&db->config->security_options);
 	}
 	return rc;
 }
@@ -283,7 +386,13 @@ static void security__module_cleanup_single(struct mosquitto__security_options *
 
 	for(i=0; i<opts->auth_plugin_config_count; i++){
 		/* Run plugin cleanup function */
-		if(opts->auth_plugin_configs[i].plugin.version == 3){
+		if(opts->auth_plugin_configs[i].plugin.version == 4){
+			opts->auth_plugin_configs[i].plugin.plugin_cleanup_v4(
+					opts->auth_plugin_configs[i].plugin.user_data,
+					opts->auth_plugin_configs[i].options,
+					opts->auth_plugin_configs[i].option_count);
+
+		}else if(opts->auth_plugin_configs[i].plugin.version == 3){
 			opts->auth_plugin_configs[i].plugin.plugin_cleanup_v3(
 					opts->auth_plugin_configs[i].plugin.user_data,
 					opts->auth_plugin_configs[i].options,
@@ -326,7 +435,14 @@ static int security__init_single(struct mosquitto__security_options *opts, bool
 	int rc;
 
 	for(i=0; i<opts->auth_plugin_config_count; i++){
-		if(opts->auth_plugin_configs[i].plugin.version == 3){
+		if(opts->auth_plugin_configs[i].plugin.version == 4){
+			rc = opts->auth_plugin_configs[i].plugin.security_init_v4(
+					opts->auth_plugin_configs[i].plugin.user_data,
+					opts->auth_plugin_configs[i].options,
+					opts->auth_plugin_configs[i].option_count,
+					reload);
+
+		}else if(opts->auth_plugin_configs[i].plugin.version == 3){
 			rc = opts->auth_plugin_configs[i].plugin.security_init_v3(
 					opts->auth_plugin_configs[i].plugin.user_data,
 					opts->auth_plugin_configs[i].options,
@@ -385,7 +501,14 @@ static int security__cleanup_single(struct mosquitto__security_options *opts, bo
 	int rc;
 
 	for(i=0; i<opts->auth_plugin_config_count; i++){
-		if(opts->auth_plugin_configs[i].plugin.version == 3){
+		if(opts->auth_plugin_configs[i].plugin.version == 4){
+			rc = opts->auth_plugin_configs[i].plugin.security_cleanup_v4(
+					opts->auth_plugin_configs[i].plugin.user_data,
+					opts->auth_plugin_configs[i].options,
+					opts->auth_plugin_configs[i].option_count,
+					reload);
+
+		}else if(opts->auth_plugin_configs[i].plugin.version == 3){
 			rc = opts->auth_plugin_configs[i].plugin.security_cleanup_v3(
 					opts->auth_plugin_configs[i].plugin.user_data,
 					opts->auth_plugin_configs[i].options,
@@ -450,7 +573,9 @@ static int acl__check_single(struct mosquitto__auth_plugin_config *auth_plugin,
 		}
 	}
 
-	if(auth_plugin->plugin.version == 3){
+	if(auth_plugin->plugin.version == 4){
+		return auth_plugin->plugin.acl_check_v4(auth_plugin->plugin.user_data, access, context, msg);
+	}else if(auth_plugin->plugin.version == 3){
 		return auth_plugin->plugin.acl_check_v3(auth_plugin->plugin.user_data, access, context, msg);
 	}else if(auth_plugin->plugin.version == 2){
 		if(access == MOSQ_ACL_SUBSCRIBE){
@@ -463,6 +588,40 @@ static int acl__check_single(struct mosquitto__auth_plugin_config *auth_plugin,
 }
 
 
+static int acl__check_dollar(const char *topic, int access)
+{
+	int rc;
+	bool match = false;
+
+	if(topic[0] != '$') return MOSQ_ERR_SUCCESS;
+
+	if(!strncmp(topic, "$SYS", 4)){
+		if(access == MOSQ_ACL_WRITE){
+			/* Potentially allow write access for bridge status, otherwise explicitly deny. */
+			rc = mosquitto_topic_matches_sub("$SYS/broker/connection/+/state", topic, &match);
+			if(rc == MOSQ_ERR_SUCCESS && match == true){
+				return MOSQ_ERR_SUCCESS;
+			}else{
+				return MOSQ_ERR_ACL_DENIED;
+			}
+		}else{
+			return MOSQ_ERR_SUCCESS;
+		}
+	}else if(!strncmp(topic, "$share", 6)){
+		/* Only allow sub/unsub to shared subscriptions */
+		if(access == MOSQ_ACL_SUBSCRIBE){
+		//FIXME if(access == MOSQ_ACL_SUBSCRIBE || access == MOSQ_ACL_UNSUBSCRIBE){
+			return MOSQ_ERR_SUCCESS;
+		}else{
+			return MOSQ_ERR_ACL_DENIED;
+		}
+	}else{
+		/* This is an unknown $ topic, for the moment just defer to actual tests. */
+		return MOSQ_ERR_SUCCESS;
+	}
+}
+
+
 int mosquitto_acl_check(struct mosquitto_db *db, struct mosquitto *context, const char *topic, long payloadlen, void* payload, int qos, bool retain, int access)
 {
 	int rc;
@@ -474,6 +633,9 @@ int mosquitto_acl_check(struct mosquitto_db *db, struct mosquitto *context, cons
 		return MOSQ_ERR_ACL_DENIED;
 	}
 
+	rc = acl__check_dollar(topic, access);
+	if(rc) return rc;
+
 	rc = mosquitto_acl_check_default(db, context, topic, access);
 	if(rc != MOSQ_ERR_PLUGIN_DEFER){
 		return rc;
@@ -532,7 +694,16 @@ int mosquitto_unpwd_check(struct mosquitto_db *db, struct mosquitto *context, co
 
 	rc = MOSQ_ERR_SUCCESS;
 	for(i=0; i<opts->auth_plugin_config_count; i++){
-		if(opts->auth_plugin_configs[i].plugin.version == 3){
+		if(opts->auth_plugin_configs[i].plugin.version == 4 
+				&& opts->auth_plugin_configs[i].plugin.unpwd_check_v4){
+
+			rc = opts->auth_plugin_configs[i].plugin.unpwd_check_v4(
+					opts->auth_plugin_configs[i].plugin.user_data,
+					context,
+					username,
+					password);
+
+		}else if(opts->auth_plugin_configs[i].plugin.version == 3){
 			rc = opts->auth_plugin_configs[i].plugin.unpwd_check_v3(
 					opts->auth_plugin_configs[i].plugin.user_data,
 					context,
@@ -581,7 +752,18 @@ int mosquitto_psk_key_get(struct mosquitto_db *db, struct mosquitto *context, co
 	}
 
 	for(i=0; i<opts->auth_plugin_config_count; i++){
-		if(opts->auth_plugin_configs[i].plugin.version == 3){
+		if(opts->auth_plugin_configs[i].plugin.version == 3
+				&& opts->auth_plugin_configs[i].plugin.psk_key_get_v4){
+
+			rc = opts->auth_plugin_configs[i].plugin.psk_key_get_v4(
+					opts->auth_plugin_configs[i].plugin.user_data,
+					context,
+					hint,
+					identity,
+					key,
+					max_key_len);
+
+		}else if(opts->auth_plugin_configs[i].plugin.version == 3){
 			rc = opts->auth_plugin_configs[i].plugin.psk_key_get_v3(
 					opts->auth_plugin_configs[i].plugin.user_data,
 					context,
@@ -613,3 +795,85 @@ int mosquitto_psk_key_get(struct mosquitto_db *db, struct mosquitto *context, co
 	return rc;
 }
 
+
+int mosquitto_security_auth_start(struct mosquitto_db *db, struct mosquitto *context, bool reauth, const void *data_in, uint16_t data_in_len, void **data_out, uint16_t *data_out_len)
+{
+	int rc = MOSQ_ERR_PLUGIN_DEFER;
+	int i;
+	struct mosquitto__security_options *opts;
+
+	if(!context || !context->listener || !context->auth_method) return MOSQ_ERR_INVAL;
+	if(!data_out || !data_out_len) return MOSQ_ERR_INVAL;
+
+	if(db->config->per_listener_settings){
+		opts = &context->listener->security_options;
+	}else{
+		opts = &db->config->security_options;
+	}
+
+	for(i=0; i<opts->auth_plugin_config_count; i++){
+		if(opts->auth_plugin_configs[i].plugin.auth_start_v4){
+			*data_out = NULL;
+			*data_out_len = 0;
+
+			rc = opts->auth_plugin_configs[i].plugin.auth_start_v4(
+					opts->auth_plugin_configs[i].plugin.user_data,
+					context,
+					context->auth_method,
+					reauth,
+					data_in, data_in_len,
+					data_out, data_out_len);
+
+			if(rc == MOSQ_ERR_SUCCESS){
+				return MOSQ_ERR_SUCCESS;
+			}else if(rc == MOSQ_ERR_AUTH_CONTINUE){
+				return MOSQ_ERR_AUTH_CONTINUE;
+			}else if(rc != MOSQ_ERR_NOT_SUPPORTED){
+				return rc;
+			}
+		}
+	}
+
+	return MOSQ_ERR_NOT_SUPPORTED;
+}
+
+
+int mosquitto_security_auth_continue(struct mosquitto_db *db, struct mosquitto *context, const void *data_in, uint16_t data_in_len, void **data_out, uint16_t *data_out_len)
+{
+	int rc = MOSQ_ERR_PLUGIN_DEFER;
+	int i;
+	struct mosquitto__security_options *opts;
+
+	if(!context || !context->listener || !context->auth_method) return MOSQ_ERR_INVAL;
+	if(!data_out || !data_out_len) return MOSQ_ERR_INVAL;
+
+	if(db->config->per_listener_settings){
+		opts = &context->listener->security_options;
+	}else{
+		opts = &db->config->security_options;
+	}
+
+	for(i=0; i<opts->auth_plugin_config_count; i++){
+		if(opts->auth_plugin_configs[i].plugin.auth_start_v4){
+			*data_out = NULL;
+			*data_out_len = 0;
+
+			rc = opts->auth_plugin_configs[i].plugin.auth_continue_v4(
+					opts->auth_plugin_configs[i].plugin.user_data,
+					context,
+					context->auth_method,
+					data_in, data_in_len,
+					data_out, data_out_len);
+
+			if(rc == MOSQ_ERR_SUCCESS){
+				return MOSQ_ERR_SUCCESS;
+			}else if(rc == MOSQ_ERR_AUTH_CONTINUE){
+				return MOSQ_ERR_AUTH_CONTINUE;
+			}else if(rc != MOSQ_ERR_NOT_SUPPORTED){
+				return rc;
+			}
+		}
+	}
+
+	return MOSQ_ERR_NOT_SUPPORTED;
+}
