@@ -1,5 +1,5 @@
 /*
-Copyright (c) 2009-2018 Roger Light <roger@atchoo.org>
+Copyright (c) 2009-2019 Roger Light <roger@atchoo.org>
 
 All rights reserved. This program and the accompanying materials
 are made available under the terms of the Eclipse Public License v1.0
@@ -52,18 +52,18 @@ void sys_tree__init(struct mosquitto_db *db)
 
 	/* Set static $SYS messages */
 	snprintf(buf, 64, "mosquitto version %s", VERSION);
-	db__messages_easy_queue(db, NULL, "$SYS/broker/version", SYS_TREE_QOS, strlen(buf), buf, 1);
+	db__messages_easy_queue(db, NULL, "$SYS/broker/version", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 }
 
 static void sys_tree__update_clients(struct mosquitto_db *db, char *buf)
 {
-	static unsigned int client_count = -1;
+	static int client_count = -1;
 	static int clients_expired = -1;
-	static unsigned int client_max = 0;
+	static int client_max = 0;
 	static int disconnected_count = -1;
 	static int connected_count = -1;
 
-	unsigned int count_total, count_by_sock;
+	int count_total, count_by_sock;
 
 	count_total = HASH_CNT(hh_id, db->contexts_by_id);
 	count_by_sock = HASH_CNT(hh_sock, db->contexts_by_sock);
@@ -71,12 +71,12 @@ static void sys_tree__update_clients(struct mosquitto_db *db, char *buf)
 	if(client_count != count_total){
 		client_count = count_total;
 		snprintf(buf, BUFLEN, "%d", client_count);
-		db__messages_easy_queue(db, NULL, "$SYS/broker/clients/total", SYS_TREE_QOS, strlen(buf), buf, 1);
+		db__messages_easy_queue(db, NULL, "$SYS/broker/clients/total", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 
 		if(client_count > client_max){
 			client_max = client_count;
 			snprintf(buf, BUFLEN, "%d", client_max);
-			db__messages_easy_queue(db, NULL, "$SYS/broker/clients/maximum", SYS_TREE_QOS, strlen(buf), buf, 1);
+			db__messages_easy_queue(db, NULL, "$SYS/broker/clients/maximum", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 		}
 	}
 
@@ -90,19 +90,19 @@ static void sys_tree__update_clients(struct mosquitto_db *db, char *buf)
 			disconnected_count = 0;
 		}
 		snprintf(buf, BUFLEN, "%d", disconnected_count);
-		db__messages_easy_queue(db, NULL, "$SYS/broker/clients/inactive", SYS_TREE_QOS, strlen(buf), buf, 1);
-		db__messages_easy_queue(db, NULL, "$SYS/broker/clients/disconnected", SYS_TREE_QOS, strlen(buf), buf, 1);
+		db__messages_easy_queue(db, NULL, "$SYS/broker/clients/inactive", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
+		db__messages_easy_queue(db, NULL, "$SYS/broker/clients/disconnected", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 	}
 	if(connected_count != count_by_sock){
 		connected_count = count_by_sock;
 		snprintf(buf, BUFLEN, "%d", connected_count);
-		db__messages_easy_queue(db, NULL, "$SYS/broker/clients/active", SYS_TREE_QOS, strlen(buf), buf, 1);
-		db__messages_easy_queue(db, NULL, "$SYS/broker/clients/connected", SYS_TREE_QOS, strlen(buf), buf, 1);
+		db__messages_easy_queue(db, NULL, "$SYS/broker/clients/active", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
+		db__messages_easy_queue(db, NULL, "$SYS/broker/clients/connected", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 	}
 	if(g_clients_expired != clients_expired){
 		clients_expired = g_clients_expired;
 		snprintf(buf, BUFLEN, "%d", clients_expired);
-		db__messages_easy_queue(db, NULL, "$SYS/broker/clients/expired", SYS_TREE_QOS, strlen(buf), buf, 1);
+		db__messages_easy_queue(db, NULL, "$SYS/broker/clients/expired", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 	}
 }
 
@@ -117,13 +117,13 @@ static void sys_tree__update_memory(struct mosquitto_db *db, char *buf)
 	if(current_heap != value_ul){
 		current_heap = value_ul;
 		snprintf(buf, BUFLEN, "%lu", current_heap);
-		db__messages_easy_queue(db, NULL, "$SYS/broker/heap/current", SYS_TREE_QOS, strlen(buf), buf, 1);
+		db__messages_easy_queue(db, NULL, "$SYS/broker/heap/current", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 	}
 	value_ul =mosquitto__max_memory_used();
 	if(max_heap != value_ul){
 		max_heap = value_ul;
 		snprintf(buf, BUFLEN, "%lu", max_heap);
-		db__messages_easy_queue(db, NULL, "$SYS/broker/heap/maximum", SYS_TREE_QOS, strlen(buf), buf, 1);
+		db__messages_easy_queue(db, NULL, "$SYS/broker/heap/maximum", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 	}
 }
 #endif
@@ -135,12 +135,12 @@ static void calc_load(struct mosquitto_db *db, char *buf, const char *topic, boo
 	if (initial) {
 		new_value = *current;
 		snprintf(buf, BUFLEN, "%.2f", new_value);
-		db__messages_easy_queue(db, NULL, topic, SYS_TREE_QOS, strlen(buf), buf, 1);
+		db__messages_easy_queue(db, NULL, topic, SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 	} else {
 		new_value = interval + exponent*((*current) - interval);
 		if(fabs(new_value - (*current)) >= 0.01){
 			snprintf(buf, BUFLEN, "%.2f", new_value);
-			db__messages_easy_queue(db, NULL, topic, SYS_TREE_QOS, strlen(buf), buf, 1);
+			db__messages_easy_queue(db, NULL, topic, SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 		}
 	}
 	(*current) = new_value;
@@ -171,6 +171,7 @@ void sys_tree__update(struct mosquitto_db *db, int interval, time_t start_time)
 	static unsigned long long pub_bytes_received = -1;
 	static unsigned long long pub_bytes_sent = -1;
 	static int subscription_count = -1;
+	static int shared_subscription_count = -1;
 	static int retained_count = -1;
 
 	static double msgs_received_load1 = 0;
@@ -218,7 +219,7 @@ void sys_tree__update(struct mosquitto_db *db, int interval, time_t start_time)
 	if(interval && now - interval > last_update){
 		uptime = now - start_time;
 		snprintf(buf, BUFLEN, "%d seconds", (int)uptime);
-		db__messages_easy_queue(db, NULL, "$SYS/broker/uptime", SYS_TREE_QOS, strlen(buf), buf, 1);
+		db__messages_easy_queue(db, NULL, "$SYS/broker/uptime", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 
 		sys_tree__update_clients(db, buf);
 		bool initial_publish = false;
@@ -287,26 +288,32 @@ void sys_tree__update(struct mosquitto_db *db, int interval, time_t start_time)
 		if(db->msg_store_count != msg_store_count){
 			msg_store_count = db->msg_store_count;
 			snprintf(buf, BUFLEN, "%d", msg_store_count);
-			db__messages_easy_queue(db, NULL, "$SYS/broker/messages/stored", SYS_TREE_QOS, strlen(buf), buf, 1);
-			db__messages_easy_queue(db, NULL, "$SYS/broker/store/messages/count", SYS_TREE_QOS, strlen(buf), buf, 1);
+			db__messages_easy_queue(db, NULL, "$SYS/broker/messages/stored", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
+			db__messages_easy_queue(db, NULL, "$SYS/broker/store/messages/count", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 		}
 
 		if (db->msg_store_bytes != msg_store_bytes){
 			msg_store_bytes = db->msg_store_bytes;
 			snprintf(buf, BUFLEN, "%lu", msg_store_bytes);
-			db__messages_easy_queue(db, NULL, "$SYS/broker/store/messages/bytes", SYS_TREE_QOS, strlen(buf), buf, 1);
+			db__messages_easy_queue(db, NULL, "$SYS/broker/store/messages/bytes", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 		}
 
 		if(db->subscription_count != subscription_count){
 			subscription_count = db->subscription_count;
 			snprintf(buf, BUFLEN, "%d", subscription_count);
-			db__messages_easy_queue(db, NULL, "$SYS/broker/subscriptions/count", SYS_TREE_QOS, strlen(buf), buf, 1);
+			db__messages_easy_queue(db, NULL, "$SYS/broker/subscriptions/count", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
+		}
+
+		if(db->shared_subscription_count != shared_subscription_count){
+			shared_subscription_count = db->shared_subscription_count;
+			snprintf(buf, BUFLEN, "%d", shared_subscription_count);
+			db__messages_easy_queue(db, NULL, "$SYS/broker/shared_subscriptions/count", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 		}
 
 		if(db->retained_count != retained_count){
 			retained_count = db->retained_count;
 			snprintf(buf, BUFLEN, "%d", retained_count);
-			db__messages_easy_queue(db, NULL, "$SYS/broker/retained messages/count", SYS_TREE_QOS, strlen(buf), buf, 1);
+			db__messages_easy_queue(db, NULL, "$SYS/broker/retained messages/count", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 		}
 
 #ifdef REAL_WITH_MEMORY_TRACKING
@@ -316,55 +323,55 @@ void sys_tree__update(struct mosquitto_db *db, int interval, time_t start_time)
 		if(msgs_received != g_msgs_received){
 			msgs_received = g_msgs_received;
 			snprintf(buf, BUFLEN, "%lu", msgs_received);
-			db__messages_easy_queue(db, NULL, "$SYS/broker/messages/received", SYS_TREE_QOS, strlen(buf), buf, 1);
+			db__messages_easy_queue(db, NULL, "$SYS/broker/messages/received", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 		}
 		
 		if(msgs_sent != g_msgs_sent){
 			msgs_sent = g_msgs_sent;
 			snprintf(buf, BUFLEN, "%lu", msgs_sent);
-			db__messages_easy_queue(db, NULL, "$SYS/broker/messages/sent", SYS_TREE_QOS, strlen(buf), buf, 1);
+			db__messages_easy_queue(db, NULL, "$SYS/broker/messages/sent", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 		}
 
 		if(publish_dropped != g_msgs_dropped){
 			publish_dropped = g_msgs_dropped;
 			snprintf(buf, BUFLEN, "%lu", publish_dropped);
-			db__messages_easy_queue(db, NULL, "$SYS/broker/publish/messages/dropped", SYS_TREE_QOS, strlen(buf), buf, 1);
+			db__messages_easy_queue(db, NULL, "$SYS/broker/publish/messages/dropped", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 		}
 
 		if(pub_msgs_received != g_pub_msgs_received){
 			pub_msgs_received = g_pub_msgs_received;
 			snprintf(buf, BUFLEN, "%lu", pub_msgs_received);
-			db__messages_easy_queue(db, NULL, "$SYS/broker/publish/messages/received", SYS_TREE_QOS, strlen(buf), buf, 1);
+			db__messages_easy_queue(db, NULL, "$SYS/broker/publish/messages/received", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 		}
 		
 		if(pub_msgs_sent != g_pub_msgs_sent){
 			pub_msgs_sent = g_pub_msgs_sent;
 			snprintf(buf, BUFLEN, "%lu", pub_msgs_sent);
-			db__messages_easy_queue(db, NULL, "$SYS/broker/publish/messages/sent", SYS_TREE_QOS, strlen(buf), buf, 1);
+			db__messages_easy_queue(db, NULL, "$SYS/broker/publish/messages/sent", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 		}
 
 		if(bytes_received != g_bytes_received){
 			bytes_received = g_bytes_received;
 			snprintf(buf, BUFLEN, "%llu", bytes_received);
-			db__messages_easy_queue(db, NULL, "$SYS/broker/bytes/received", SYS_TREE_QOS, strlen(buf), buf, 1);
+			db__messages_easy_queue(db, NULL, "$SYS/broker/bytes/received", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 		}
 		
 		if(bytes_sent != g_bytes_sent){
 			bytes_sent = g_bytes_sent;
 			snprintf(buf, BUFLEN, "%llu", bytes_sent);
-			db__messages_easy_queue(db, NULL, "$SYS/broker/bytes/sent", SYS_TREE_QOS, strlen(buf), buf, 1);
+			db__messages_easy_queue(db, NULL, "$SYS/broker/bytes/sent", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 		}
 		
 		if(pub_bytes_received != g_pub_bytes_received){
 			pub_bytes_received = g_pub_bytes_received;
 			snprintf(buf, BUFLEN, "%llu", pub_bytes_received);
-			db__messages_easy_queue(db, NULL, "$SYS/broker/publish/bytes/received", SYS_TREE_QOS, strlen(buf), buf, 1);
+			db__messages_easy_queue(db, NULL, "$SYS/broker/publish/bytes/received", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 		}
 
 		if(pub_bytes_sent != g_pub_bytes_sent){
 			pub_bytes_sent = g_pub_bytes_sent;
 			snprintf(buf, BUFLEN, "%llu", pub_bytes_sent);
-			db__messages_easy_queue(db, NULL, "$SYS/broker/publish/bytes/sent", SYS_TREE_QOS, strlen(buf), buf, 1);
+			db__messages_easy_queue(db, NULL, "$SYS/broker/publish/bytes/sent", SYS_TREE_QOS, strlen(buf), buf, 1, 60, NULL);
 		}
 
 		last_update = mosquitto_time();
